#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "wsl_types.h"
#include "wsl_errs.h"
#include "code.h"
#include "code.cpp"
#include "decode.cpp"

void Quit(int err, char* mesB)
{
  if (err != WC_OK)
    printf("\n%s%s\n", WC_ERRORS[err], mesB);
  exit(err != WC_OK);
}

void ShowHelp()
{
  Quit(WC_HELP,
  "Yet another LZW implementation.\n\n\
Usage: WSL <e|d> inputfile outputfile [<switches>]\n\n\
  <Switches>\n\
    -bN     Set maximum blocksize in kb  (def: -b64, min: -b1, max: -b16384)\n\
    -dN     Set dictionary size in kb (def: -d32, min: -d1, max: -d64)\n\
    -lN     Set search depth (def: -l4, min: -l1, max: -l50)\n\
Example: WSL e FILENAME_IN FILENAME_OUT -b128d32l1\n\
  (set blocksize to 128kb and perform compression with 32kb dictionary\n\
   and minimum search depth)");
}

int ParseParams(int argc, char* argv[], PSets s)
{
  memset(s, 0, sizeof(TSets));
  s->block_size = DEFBLOCKSIZE;
  s->dic_size = DEFDICSIZE;
  s->sl = DEFSL;
  
  int i, j;
  uint32 n;
  for (i = 4; i < argc; i++)
    if (argv[i][0] == '-')
    {
      int k = 1;
      for (; argv[i][k]; k++)
        switch (argv[i][k])
        {
          case 'b' : 
            n = 0;
            for (j = k + 1; argv[i][j]; j++)
            {
              if ((argv[i][j] < '0') || (argv[i][j] > '9')) break;
              n = n*10 + argv[i][j] - '0';
            }
            if (j == k + 1) return i;
            k = j - 1;
            n *= 1024;
            if (n < MINBLOCKSIZE) return i;
            if (n > MAXBLOCKSIZE) return i;
            s->block_size = n;
            break;
          case 'd' : 
            n = 0;
            for (j = k + 1; argv[i][j]; j++)
            {
              if ((argv[i][j] < '0') || (argv[i][j] > '9')) break;
              n = n*10 + argv[i][j] - '0';
            }
            if (j == k + 1) return i;
            k = j - 1;
            if (n < MINDICSIZE) return i;
            if (n > MAXDICSIZE) return i;
            s->dic_size = n;
            break;
          case 'l' : 
            n = 0;
            for (j = k + 1; argv[i][j]; j++)
            {
              if ((argv[i][j] < '0') || (argv[i][j] > '9')) break;
              n = n*10 + argv[i][j] - '0';
            }
            if (j == k + 1) return i;
            k = j - 1;
            if (n < MINSL) return i;
            if (n > MAXSL) return i;
            s->sl = n;
            break;
          default  : return i;
        }
      if (k == 1) return i;
    }
    else
      return i;
  return 0;
}

void Compress(PSets s)
{
  uint32 ins, ots, res, fs, cfs;
  PBuf ibuf, obuf;
  ibuf = (PBuf)malloc(s->block_size + BPLUS);
  obuf = (PBuf)malloc(s->block_size + BPLUS);  
  
  char head[] = HEADER;

  FILE* fin  = fopen(s->infile, "rb");     
  if (fin == NULL)  Quit(WC_ERR_INPUT, s->infile);
  FILE* fout = fopen(s->outfile, "wb");
  if (fout == NULL) Quit(WC_ERR_OUTPUT, s->outfile);
  
  fseek(fin, 0, SEEK_END);
  fs = ftell(fin) + 1;
  rewind(fin);

  if (InitCode(s->dic_size, s->sl) == 0) 
    Quit(WC_ERR_OUTMEM, "");

  printf("\nCompressing %s to %s ", s->infile, s->outfile);
  cfs = 5 + strlen(head);

  res = fwrite(head, 1, strlen(head), fout);
  if (res != strlen(head)) 
    Quit(WC_ERR_OUTPUT, s->outfile);
  res = fwrite(&s->block_size, 1, 4, fout);
  if (res != 4) 
    Quit(WC_ERR_OUTPUT, s->outfile);
  res = fwrite(&s->dic_size, 1, 1, fout);
  if (res != 1) 
    Quit(WC_ERR_OUTPUT, s->outfile);
  while (1)
  {
    ins = fread(ibuf, 1, s->block_size, fin);
    if (ins == 0) break;
    Code(&ibuf, ins, &obuf, &ots);
    res = fwrite(&ots, 1, 4, fout);
    if (res != 4)
      Quit(WC_ERR_OUTPUT, s->outfile);
    res = fwrite(obuf, 1, ots, fout);
    if (res != ots)
      Quit(WC_ERR_OUTPUT, s->outfile);
    cfs += res + 4;
  }
  DoneCode();
  printf(" Done, %ld -> %ld, %.2f%%, %.2f bpb\n", fs - 1, cfs,
    (double(cfs)/fs)*100, (double(cfs)/fs)*8);

  fclose(fin);
  fclose(fout);
  free(ibuf);
  free(obuf);
}

void Decompress(PSets s)
{
  uint32 bs, fs, rfs, res, ins, ots;
  byte ds;
  PBuf ibuf, obuf;
  char head[] = HEADER;
  char bh[]   = HEADER;

  FILE* fin  = fopen(s->infile, "rb");     
  if (fin == NULL)  Quit(WC_ERR_INPUT, s->infile);
  FILE* fout = fopen(s->outfile, "wb");
  if (fout == NULL) Quit(WC_ERR_OUTPUT, s->outfile);  

  fseek(fin, 0, SEEK_END);
  fs = ftell(fin);
  rewind(fin);

  res = fread(bh, 1, strlen(head), fin);
  if (res != strlen(head) || strcmp(head, bh) != 0) 
    Quit(WC_ERR_HEADER, head);
  res = fread(&bs, 1, 4, fin);
  if (res != 4) 
    Quit(WC_ERR_INPUT, s->infile);
  res = fread(&ds, 1, 1, fin);
  if (res != 1) 
    Quit(WC_ERR_INPUT, s->infile);

  ibuf = (PBuf)malloc(bs + BPLUS);
  obuf = (PBuf)malloc(bs + BPLUS);  
  
  if (InitDeCode(ds) == 0) Quit(WC_ERR_OUTMEM, "");
  
  printf("\nDecompressing %s to %s ", s->infile, s->outfile);
 
  rfs = 0;
  while (1)
  {
    res = fread(&ins, 1, 4, fin);
    if (res == 0) break;
    if (res != 4)
      Quit(WC_ERR_INPUT, s->infile);
    res = fread(ibuf, 1, ins, fin);
    if (res == 0) break;
    DeCode(&ibuf, ins, &obuf, &ots); 
    res = fwrite(obuf, 1, ots, fout);
    if (res != ots)
      Quit(WC_ERR_OUTPUT, s->outfile);
    rfs += res;
  }
  DoneDeCode();
  printf(" Done, %ld -> %ld\n", fs, rfs);

  fclose(fin);
  fclose(fout);
  free(ibuf);
  free(obuf);
}

int main(int argc, char* argv[])
{
  TSets sets;
  
  printf(CAPTION);

  if (argc < 4) ShowHelp();

  if ((toupper(argv[1][0]) != 'E' &&
       toupper(argv[1][0]) != 'D') || argv[1][1]) 
    Quit(WC_ERR_CMD, argv[1]);

  if (int pp = ParseParams(argc, argv, &sets)) 
    Quit(WC_ERR_SWITCH, argv[pp]);

  sets.infile  = argv[2];
  sets.outfile = argv[3];
  
  uint32 time = clock();
  if (toupper(argv[1][0]) == 'E')
    Compress(&sets);  
  else
    Decompress(&sets);
  printf("Time: %.2f\n", (double)(clock() - time)/1000);
  
  Quit(WC_OK, "");
  return 0;
}
